import asyncio
import redis
import aioredis
import logging

from .serialization import serialize_task, serialize, deserialize
from .status import TaskStatus
from .errors import InvalidTaskStatus
from .util import now


class TaskTrackerBase(object):
    def __init__(self, *args, logger=None):
        from .base import config
        self.config = config
        self.logger = logger or logging.getLogger()

    def key_name(self, task_id):
        return ":".join([self.config.status_prefix, str(task_id)])

    # def mark_status(self, task_id, new_status):
    #     if new_status not in TaskStatus:
    #         raise InvalidTaskStatus("new_status must be type of TaskStatus")

    def complete(self, task_id, result=None, status=TaskStatus.SUCCESS, excinfo=None):
        key_name = self.key_name(task_id)
        body = {'status': serialize(status)}
        if result:
            body['result'] = serialize(result)
        if excinfo:
            body['excinfo'] = serialize(excinfo)
        return (key_name, body)

    def schedule(self, task_id, func, args, kwargs):
        raise NotImplementedError("implement in subclass")

    def get_task_by_uuid(self, task_id):
        raise NotImplementedError()

    def empty(self):
        raise NotImplementedError()


class TaskTracker(TaskTrackerBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.redis = redis.StrictRedis(**self.config.redis_opts)

    def mark_status(self, task_id, new_status):
        redis_key = self.key_name(task_id)
        self.redis.expire(redis_key, self.config.result_expiry_time)
        self.redis.hset(redis_key, 'status', new_status)

    def complete(self, *args, **kwargs):
        key_name, body = super().complete(*args, **kwargs)
        self.redis.hmset(key_name, body)

    def schedule(self, task_id, func, args, kwargs):
        serialized = serialize_task(task_id, func, args, kwargs)
        key_name = self.key_name(task_id)
        with self.redis.pipeline() as pipe:
            pipe.lpush(self.config.queue_name, serialized)
            pipe.hmset(key_name, {
                'status': serialize(TaskStatus.INIT),
                'submitted': serialize(now()),
                'result': serialize(None),
                'body': serialized
            })
            pipe.expire(key_name, self.config.result_expiry_time)
            pipe.execute()

    def get_task_by_uuid(self, task_id):
        return self.redis.hgetall(self.key_name(task_id))

    def empty(self):
        return self.redis.llen(self.config.queue_name) == 0

    def get(self, block=False, timeout=None):
        if block:
            item = self.redis.brpop(self.config.queue_name, timeout=timeout or 0)
        else:
            item = self.redis.rpop(self.config.queue_name)

        if item:
            task_id, func_name, args, kwargs = deserialize(item)
            return (task_id, func_name, args, kwargs)
        return None


class AsyncTaskTracker(TaskTrackerBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        result = asyncio.get_event_loop().run_until_complete(asyncio.gather(self._install_redis()))
        self.redis = aioredis.Redis(result[0])

    async def _install_redis(self):
        return await aioredis.create_connection(self.config.redis_url)

    async def mark_status(self, task_id, new_status):
        redis_key = self.key_name(task_id)
        await self.redis.expire(redis_key, self.config.result_expiry_time)
        await self.redis.hset(redis_key, 'status', serialize(new_status))
        return 1

    async def complete(self, *args, **kwargs):
        key_name, body = super().complete(*args, **kwargs)
        # import pdb; pdb.set_trace()
        await self.redis.hmset_dict(key_name, body)

    async def schedule(self, task_id, func, args, kwargs):
        serialized = serialize_task(task_id, func, args, kwargs)
        key_name = self.key_name(task_id)

        pipe = self.redis.pipeline()
        pipe.lpush(self.config.queue_name, serialized)
        pipe.hmset_dict(key_name, {
            'status': serialize(TaskStatus.INIT),
            'submitted': serialize(now()),
            'result': serialize(None),
            'body': serialized
        })
        pipe.expire(key_name, self.config.result_expiry_time)
        await pipe.execute()

    async def get_task_by_uuid(self, task_id):
        return await self.redis.hgetall(self.key_name(task_id))

    async def empty(self):
        return (await self.redis.llen(self.config.queue_name)) == 0

    async def get(self, block=False):
        if block:
            item = await self.redis.brpop(self.config.queue_name, timeout=10)
        else:
            item = await self.redis.rpop(self.config.queue_name)
            while item is None:
                await asyncio.sleep(1)
                item = await self.redis.rpop(self.config.queue_name)

        task_id, func_name, args, kwargs = deserialize(item)
        return (task_id, func_name, args, kwargs)



import uuid
import asyncio
from urllib.parse import urlparse

from .serialization import deserialize
from .util import singleton
from .status import TaskStatus

from .tracker import TaskTracker


@singleton
class ConfigMap(object):
    def __init__(self, redis_url=None, status_prefix=None, queue_name=None, worker_count=None, result_expiry_time=3600):
        self.redis_url = redis_url
        self.status_prefix = status_prefix
        self.queue_name = queue_name
        self.worker_count = worker_count
        self.result_expiry_time = result_expiry_time

    def load(self, **config):
        for opt in ['redis_url', 'status_prefix', 'queue_name', 'worker_count', 'result_expiry_time']:
            setattr(self, opt, config.get(opt))

    @property
    def redis_opts(self):
        opts = {}
        redis_info = urlparse(self.redis_url)
        opts.update({
            'host': redis_info.hostname,
            'port': redis_info.port,
            'password': redis_info.password,
        })
        if len(redis_info.path) > 0:
            opts['db'] = redis_info.path[1:]
        return opts

config = ConfigMap.instance()


@singleton
class Registry(object):
    tasks = {}

    def task(self, func, **opts):
        """Decorator to turn a function into a task.

        Also registers the function in the task registry if it
        has not already.

        To use it, wrap your function with this decorator.

        @task
        def function_name_here():
            pass
        """
        if func.__name__ not in self.tasks:
            self.tasks[func.__name__] = dict(func=func, opts=opts, name=func.__name__)
        task_obj = Task(func=func)
        return task_obj

    def async_task(self, func, **opts):
        # TODO: preserve __name__ and others.
        opts.update({'is_async': True})
        if func.__name__ not in self.tasks:
            self.tasks[func.__name__] = dict(func=func, opts=opts)
        task_obj = Task(func=func)
        return task_obj


registry = Registry.instance()
task = registry.task
async_task = registry.async_task



class Task(object):
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.func_name = self.func.__name__
        self.task_id = None
        self.status = TaskStatus.INIT
        self.tracker = TaskTracker()
        self.result = None

    def __call__(self, *args, **kwargs):
        if asyncio.iscoroutinefunction(self.func):
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(asyncio.gather(self.func(*args, **kwargs)))
            return result[0]
        else:
            return self.func(*args, **kwargs)

    @classmethod
    def get(cls, task_id):
        attrs = TaskTracker().get_task_by_uuid(task_id)
        uuid, func_name, args, kwargs = deserialize(attrs[b'body'])
        registry_item = registry.tasks.get(func_name)

        task = cls(func=registry_item['func'])
        task.funcname = registry_item['name']
        task.status = deserialize(attrs[b'status'])
        task.result = deserialize(attrs[b'result'])
        task.submitted = deserialize(attrs[b'submitted'])

        body = deserialize(attrs[b'body'])
        task.args = body[2]
        task.kwargs = body[3]
        if b'excinfo' in attrs:
            task.excinfo = deserialize(attrs[b'excinfo'])

        task.task_id = task_id

        return task

    def refresh(self):
        task_id = self.task_id
        self = Task.get(task_id)
        return self

    async def call_async(self, *args, **kwargs):
        result = await self.func(*args, **kwargs)
        return result

    def __repr__(self):
        return r'<Task name=%s status=%s task_id=%s>' % (self.func_name, self.status, self.task_id)

    def enqueue_inline(self, *args, **kwargs):
        self.task_id = uuid.uuid4()
        self.tracker.schedule(self.task_id, self.func_name, args, kwargs)
        return self

    def enqueue_async(self, *args, **kwargs):
        pass

    enqueue = enqueue_inline



"""Basic processing worker.

You may want to tune this slightly to fit your app's specific needs.
Nonetheless, it's a good starting point.
"""

import asyncio
import logging

from .base import TaskStatus, registry, config
from .tracker import TaskTracker, AsyncTaskTracker

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class ExecutionError(Exception):
    pass


class Worker(object):
    def __init__(self, worker_count, is_async=False):
        self._redis = None
        self.worker_count = worker_count
        self.should_accept_work = True
        if is_async:
            self.tracker = AsyncTaskTracker()
        else:
            self.tracker = TaskTracker()
        self.loop = asyncio.get_event_loop()
        self.scheduler = None

    async def worker(self, number):
        while 1:
            if not self.should_accept_work:
                break
            item = await self.tracker.get()
            if item:
                result = await self.process_item(*item)
            else:
                print(f"worker {number} got nothing, sleeping")
                await asyncio.sleep(1)

    async def process_item(self, task_id, func_name, args, kwargs):
        await self.tracker.mark_status(task_id, TaskStatus.RUNNING)
        if func_name not in registry.tasks:
            logger.warn(f"Could not find task {func_name}")
        else:
            func = registry.tasks[func_name]['func']
            try:
                if asyncio.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    await asyncio.sleep(0)
                    result = func(*args, **kwargs)
            except Exception as exc:
                await self.tracker.complete(task_id, result=None, status=TaskStatus.ERROR, excinfo=exc)
            else:
                await self.tracker.complete(task_id, result=result)
                return result

    def run_forever(self):
        workers = [self.loop.create_task(self.worker(i)) for i in range(self.worker_count)]
        self.loop.run_until_complete(asyncio.gather(*workers))


if __name__ == '__main__':
    config.redis_url = "redis://localhost:6379/0"
    config.queue_name = "work_queue"
    config.status_prefix = "task_result"

    worker = Worker(4, is_async=True)
    worker.run_forever()



import pickle


class BaseSerializer(object):
    pass

class PickleSerializer(BaseSerializer):
    @staticmethod
    def serialize_task(task_id, func, args, kwargs):
        obj = (task_id, func, args, kwargs)
        return pickle.dumps(obj)

    @staticmethod
    def serialize(*args, **kwargs):
        return pickle.dumps(*args, **kwargs)

    @staticmethod
    def deserialize(obj):
        return pickle.loads(obj)

default_serializer = PickleSerializer
serialize = default_serializer.serialize
serialize_task = default_serializer.serialize_task
deserialize = default_serializer.deserialize



import re
from setuptools import setup
from os import path

project_path = path.abspath(path.dirname(__file__))


meta_file = open(path.join(project_path, "elon", "metadata.py")).read()
md = dict(re.findall(r"__([a-z]+)__\s*=\s*'([^']+)'", meta_file))

with open(path.join(project_path, 'README.md')) as f:
    long_description = f.read()


setup(
    name='elon',
    version=md['version'],
    author=md['author'],
    author_email=md['authoremail'],
    packages=['elon'],
    url="http://github.com/omarish/elon",
    license='MIT',
    description='Lightweight async job queue backed by redis.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['six>=1.11.0', 'aioredis>=1.0.0', 'redis>=2.10.5']
)
